{****************************************************************************

                      Copyright (c) 1996 by Florian Klaempfl

 ****************************************************************************}

{
  This unit implements an asmoutput class for Intel syntax with
  Intel i386+.

  History:
      26th november 1996:
         + unit derieved from attasmi3

  What's to do:
    o ASCII output
    o generate extern entries for typed constants and variables
    o generate extern entries for the internal used functions
    o complete op2str array
    o write lines numbers and file names to output file
}
unit intasmi3;

  interface

    uses
       globals,systems,errors,cobjects,aasm,i386,strings;

    type
       pi386intelasmoutput = ^ti386intelasmoutput;

       ti386intelasmoutput = object(taasmoutput)
          procedure write_to_file(outfile : pbufferedfile);virtual;
       end;

    { inits the datasegment etc with the output specific list types }
    procedure asmoutlists_init;

    { writes the asmlists to an output file }
    procedure writeasmlists(outfile : pbufferedfile);

  implementation

    const
       _reg2str : array[R_NO..R_SS] of string[4] =
          ('','eax','ecx','edx','ebx','esp','ebp','esi','edi',
           'ax','cx','dx','bx','sp','bp','si','di',
           'al','cl','dl','bl',
           '','cd','ds','es','fs','gs','ss');

       op2str : array[A_MOV..A_LAHF] of string[7] =
         ('mov','movzx','movsx','','add',
          'call','idiv','imul','jmp','lea','mul','neg','not',
          'pop','popad','push','pushad','ret','sub','xchg','xor',
          'fild','cmp','jz','inc','dec','sete','setne','setl',
          'setg','setle','setge','je','jne','jl','jg','jle','jge',
          'or','fld','fadd','fmul','fsub','fdivr','fchs','fld1',
          'fidiv','cdq','jnz','fstp','and','jno','','',
          'enter','leave','cld','movs','rep','shl','shr','bound',
          'jns','js','jo','sar','test',
          'fcom','fcomp','fcompp','fxch','faddp','fmulp','fsubrp','fdivrp',
          'fnsts','sahf','fdivp','fsubp','setc','setnc','jc','jnc',
          'ja','jae','jb','jbe','seta','setae','setb','setbe',
          'aaa','aad','aam','aas','cbw','cdq','clc','cli',
          'clts','cmc','cwd','cwde','daa','das','hlt','iret','lahf');

    function getreferencestring(const ref : treference) : string;

      var
         s : string;
         first : boolean;

      begin
         if ref.isintvalue then
           s:='$'+tostr(ref.offset)
         else
           with ref do
             begin
                first:=true;
                { have we a segment prefix ? }
                if segment<>R_DEFAULT_SEG then
                  s:=_reg2str[segment]+':['
                else s:='[';

                if assigned(symbol) then
                  begin
                     s:=s+symbol^;
                     first:=false;
                  end;
                if (base<>R_NO) then
                  begin
                     if not(first) then
                       s:=s+'+'
                     else
                       first:=false;
                     s:=s+_reg2str[base];
                  end;
                if (index<>R_NO) then
                  begin
                     if not(first) then
                       s:=s+'+'
                     else
                       first:=false;
                     s:=s+_reg2str[index];
                     if scalefactor<>0 then
                       s:=s+'*'+tostr(scalefactor);
                  end;
                if offset<0 then
                  s:=s+tostr(offset)
                else if (offset>0) then
                  s:=s+'+'+tostr(offset);
                s:=s+']';
             end;
         getreferencestring:=s;
      end;

    function getopstr(t : byte;o : pointer;s : topsize;dest : boolean) : string;

      var
         hs : string;

      begin
         case t of
            top_reg : getopstr:=_reg2str[tregister(o)];
            top_ref : begin
                         hs:=getreferencestring(preference(o)^);
                         case s of
                            S_B : hs:='byte ptr '+hs;
                            S_W : hs:='word ptr '+hs;
                            S_L : hs:='dword ptr '+hs;
                            S_BW : if dest then
                                     hs:='word ptr '+hs
                                   else
                                     hs:='byte ptr '+hs;
                            S_BL : if dest then
                                     hs:='dword ptr '+hs
                                   else
                                     hs:='byte ptr '+hs;
                            S_WL : if dest then
                                     hs:='dword ptr '+hs
                                   else
                                     hs:='word ptr '+hs;
                         end;
                         getopstr:=hs;
                      end;
            top_const : getopstr:=tostr(longint(o));
            top_symbol : begin
                            hs[0]:=char(strlen(pchar(pcsymbol(o)^.symbol)));
                            move(pchar(pcsymbol(o)^.symbol)^,hs[1],byte(hs[0]));
                            hs:='offset '+hs;
                            if pcsymbol(o)^.offset>0 then
                              hs:=hs+'+'+tostr(pcsymbol(o)^.offset)
                            else if pcsymbol(o)^.offset<0 then
                              hs:=hs+tostr(pcsymbol(o)^.offset);
                            getopstr:=hs;
                         end;
            else internalerror(10001);
         end;
      end;

    function getopstr_jmp(t : byte;o : pointer) : string;

      var
         hs : string;

      begin
         case t of
            top_reg : getopstr_jmp:=_reg2str[tregister(o)];
            top_ref : getopstr_jmp:=getreferencestring(preference(o)^);
            top_const : getopstr_jmp:=tostr(longint(o));
            top_symbol : begin
                            hs[0]:=char(strlen(pchar(pcsymbol(o)^.symbol)));
                            move(pchar(pcsymbol(o)^.symbol)^,hs[1],byte(hs[0]));
                            if pcsymbol(o)^.offset>0 then
                              hs:=hs+'+'+tostr(pcsymbol(o)^.offset)
                            else if pcsymbol(o)^.offset<0 then
                              hs:=hs+tostr(pcsymbol(o)^.offset);
                            getopstr_jmp:=hs;
                         end;
            else internalerror(10001);
         end;
      end;

{****************************************************************************
                             TI386ATTASMOUTPUT
 ****************************************************************************}

    procedure ti386intelasmoutput.write_to_file(outfile : pbufferedfile);

      var
         hp : pai;
         s : string;
         bss_offset : longint;

      begin
         bss_offset:=0;
         hp:=pai(first);
         while assigned(hp) do
           begin
              case hp^.typ of
                 ait_external : begin
                                   outfile^.write_string(#9#9'EXTRN'#9);
                                   outfile^.write_pchar(pai_external(hp)^.name);
                                   outfile^.write_string(' :BYTE');
                                end;
                 ait_datablock : begin
                                    if pai_datablock(hp)^.is_global then
                                      begin
                                         outfile^.write_string(#9#9'PUBLIC'#9);
                                         outfile^.write_pchar(pai_datablock(hp)^.name);
                                         outfile^.write_string(target_info.newline);
                                      end;
                                    if bss_offset=0 then
                                      begin
                                         outfile^.write_string(#9#9'ORG'#9'0');
                                         outfile^.write_string(target_info.newline);
                                      end;
                                    outfile^.write_pchar(pai_datablock(hp)^.name);
                                    outfile^.write_string(#9#9'LABEL'#9'BYTE');
                                    outfile^.write_string(target_info.newline);

                                    { increment BSS size }
                                    inc(bss_offset,pai_datablock(hp)^.size);
                                    outfile^.write_string(#9#9'ORG'#9+tostr(bss_offset));
                                 end;
                 ait_const_32bit : outfile^.write_string(#9#9+'DD'#9+tostr(pai_const(hp)^.value));
                 ait_const_16bit : outfile^.write_string(#9#9+'DW'#9+tostr(pai_const(hp)^.value));
                 ait_const_8bit : outfile^.write_string(#9#9+'DB'#9+tostr(pai_const(hp)^.value));
                 ait_const_symbol : begin
                                       outfile^.write_string(#9#9+'DD '#9'offset ');
                                       outfile^.write_pchar(pchar(pai_const(hp)^.value));
                                    end;
                 ait_real_64bit : outfile^.write_string(#9#9+'DQ'#9+double2str(pai_double(hp)^.value));
                 ait_string : {!!!!!};
                 ait_label : outfile^.write_string('L'+tostr(pai_label(hp)^.l)+':');
                 ait_labeled_instruction : outfile^.write_string(#9#9+op2str[pai_labeled386(hp)^.operator]+#9'L'+
                                             tostr(pai_labeled386(hp)^.lab));
                 ait_comment : {!!!!!};
                 ait_symbol : begin
                                 if pai_symbol(hp)^.is_global then
                                   begin
                                      outfile^.write_string(#9#9'PUBLIC'#9);
                                      outfile^.write_pchar(pai_symbol(hp)^.name);
                                      outfile^.write_string(target_info.newline);
                                   end;
                                 outfile^.write_pchar(pai_symbol(hp)^.name);
                                 if assigned(hp^.next) and not(pai(hp^.next)^.typ in
                                   [ait_const_32bit,ait_const_16bit,ait_const_8bit,ait_const_symbol,
                                    ait_real_64bit,ait_string]) then
                                   outfile^.write_string(':');
                              end;
                 { writes an instruction, highly table driven }
                 ait_instruction : begin
                                      if pai386(hp)^.op1t<>top_none then
                                        begin
                                           if pai386(hp)^.operator in [A_CALL] then
                                             begin
                                                s:='near ptr '+getopstr_jmp(pai386(hp)^.op1t,pai386(hp)^.op1);
                                             end
                                           else
                                             begin
                                                s:=getopstr(pai386(hp)^.op1t,pai386(hp)^.op1,pai386(hp)^.size,false);
                                                if pai386(hp)^.op2t<>top_none then
                                                  begin
                                                     s:=getopstr(pai386(hp)^.op2t,pai386(hp)^.op2,pai386(hp)^.size,true)+','+s;
                                                     if pai386(hp)^.op3t<>top_none then
                                                       s:=getopstr(pai386(hp)^.op3t,pai386(hp)^.op3,pai386(hp)^.size,false)+
                                                         ','+s;
                                                  end;
                                             end;
                                           s:=#9+s;
                                        end
                                      else
                                        s:='';
                                      s:=#9#9+op2str[pai386(hp)^.operator]+s;
                                      outfile^.write_string(s);
                                   end;
                 else internalerror(10000);
              end;
              if (hp^.typ<>ait_symbol) or (assigned(hp^.next) and not(pai(hp^.next)^.typ in
                [ait_const_32bit,ait_const_16bit,ait_const_8bit,ait_const_symbol,
                 ait_real_64bit,ait_string])) then
              outfile^.write_string(target_info.newline);
              hp:=pai(hp^.next);
           end;
      end;

    procedure asmoutlists_init;

      begin
         datasegment:=new(pi386intelasmoutput,init);
         codesegment:=new(pi386intelasmoutput,init);
         bsssegment:=new(pi386intelasmoutput,init);
         debuglist:=new(pi386intelasmoutput,init);
         externals:=new(pi386intelasmoutput,init);
      end;

    procedure writeasmlists(outfile : pbufferedfile);

      begin
         outfile^.write_string('.386p');
         outfile^.write_string(target_info.newline);

         externals^.write_to_file(outfile);

         outfile^.write_string(target_info.newline);
         outfile^.write_string('DGROUP'#9#9'GROUP'#9'_BSS,_DATA');
         outfile^.write_string(target_info.newline);
         outfile^.write_string('_TEXT'#9#9'SEGMENT'#9'BYTE PUBLIC USE32 ''CODE''');
         outfile^.write_string(target_info.newline);
         outfile^.write_string(#9#9'ASSUME'#9'CS:_TEXT,ES:DGROUP,DS:DGROUP,SS:DGROUP');
         outfile^.write_string(target_info.newline);
         codesegment^.write_to_file(outfile);
         outfile^.write_string('_TEXT'#9#9'ENDS');
         outfile^.write_string(target_info.newline);
         outfile^.write_string(target_info.newline);

         outfile^.write_string('_DATA'#9#9'SEGMENT'#9'DWORD PUBLIC USE32 ''DATA''');
         outfile^.write_string(target_info.newline);

         { write a signature to the file }
         outfile^.write_string(#9#9'DB'#9'"compiled by FPKPascal '+version_string+'\0"');
         outfile^.write_string(target_info.newline);
         outfile^.write_string(#9#9'DB'#9'"target: '+target_info.target_name+'\0"');
         outfile^.write_string(target_info.newline);

         datasegment^.write_to_file(outfile);
         outfile^.write_string('_DATA'#9#9'ENDS');
         outfile^.write_string(target_info.newline);
         outfile^.write_string(target_info.newline);

         outfile^.write_string('_BSS'#9#9'SEGMENT'#9'DWORD PUBLIC USE32 ''BSS''');
         outfile^.write_string(target_info.newline);
         bsssegment^.write_to_file(outfile);
         outfile^.write_string('_BSS'#9#9'ENDS');
         outfile^.write_string(target_info.newline);
         outfile^.write_string(target_info.newline);

         outfile^.write_string(#9#9'END');
         outfile^.write_string(target_info.newline);
      end;

end.
