{****************************************************************************

                     Copyright (c) 1996 by Florian Klaempfl

 ****************************************************************************}
{ this unit manages the make and build process }
unit unitm;

  interface

    uses
       dos,strings,cobjects,objects,globals,symtable,systems;

    type
       punit = ^tunit;

       tunit = object
          { contains the global symtable of the unit     }
          { p=nil means that the unit must be recompiled }
          symtab : punitsymtable;

          { path to the source file }
          path : pstring;
          name : pstring;
          sourcename : pstring;
          sourceext : pstring;
          next : punit;
          constructor init(const p,n : string;anext : punit);
       end;

    function searchcompile : punit;
    function readunit(const s : string) : punitsymtable;
    function searchunit(const s : string) : punit;

    var
       units : punit;

  implementation
{$I-}
{**************************************
               TUNIT
**************************************}

    constructor tunit.init(const p,n : string;anext : punit);

      begin
         symtab:=nil;
         next:=anext;
         path:=stringdup(p);
         name:=stringdup(n);
         sourcename:=nil;
         sourceext:=nil;
      end;

    function searchcompile : punit;

      var
         hp : punit;

      begin
         searchcompile:=nil;
         hp:=units;
         while assigned(hp) do
           begin
              if hp^.symtab=nil then
                begin
                   searchcompile:=hp;
                   exit;
                end;
              hp:=hp^.next;
           end;
      end;

    function searchunit(const s : string) : punit;

      var
         hp : punit;

      begin
         searchunit:=nil;
         hp:=units;
         while assigned(hp) do
           begin
              if upper(hp^.name^)=upper(s) then
                begin
                   searchunit:=hp;
                   exit;
                end;
              hp:=hp^.next;
           end;
      end;

    function readunit(const s : string) : punitsymtable;

      var
         unitfile : file;
         dummy,source_avail : boolean;
         unitfilesize : longint;
         b : byte;
         hp : punit;
         oldbuffer : prbuffer;
         oldbufferl,oldbufferp : longint;
         checksum,unit_time,source_time : longint;
         source_path : string;

      begin
         hp:=searchunit(s);
         if assigned(hp) then
           begin
              { try already to load the unit ? }
              if (hp^.symtab=nil) then
                { we can't recover from such an error }
                fatalerror(rec_unit_def)
              else
                begin
                   readunit:=hp^.symtab;
                   exit;
                end;
           end;
         oldbuffer:=rbuffer;
         oldbufferp:=bufferp;
         oldbufferl:=bufferl;
{$ifdef linux}
         from:=globals.lowercase(from);
{$endif}

         { unit isn't loaded }
         { try to load the unit }

         { search first the source }
         source_path:=search(s+target_info.sourceext,inputdir,source_avail);
         if not(dummy) then
           source_path:=search(s+target_info.sourceext,unitsearchpath,source_avail);

         { search the directory of the input file }
         unit_path:=search(s+target_info.unitext,inputdir,dummy);
         if not(dummy) then
           unitpath:=search(s+target_info.unitext,unitsearchpath,dummy);

         if not(dummy) then
           begin
              { we must in any case compile the unit }
              { if we don't find the PPU file        }
              if source_avail then
                begin
                   { source is available          }
                   { so we could compile the unit }
                   units:=new(punit,init(source_path,s,units));
                   units^.sourcename:=stringdup(s);
                   units^.sourceext:=stringdup(target_info.sourceext);
                   readunit:=nil;
                   exit;
                end
              else
                begin
                   exterror:=strpnew(s+target_info.unitext);
                   { I think, it's seamless to try to recover from }
                   { this error                                    }
                   fatalerror(unit_not_found);
                end;
           end
         else
           begin
              { if we have a PPU and a PP file: }
              if source_avail then
                begin
                   { get file times }
                   assign(unitfile,unit_path+s+target_info.unitext);
                   getftime(unitfile,unit_time);
                   assign(unitfile,source_path+s+target_info.sourceext);
                   getftime(unitfile,source_time);

                   { compile the source:         }
                   { if there is a build command }
                   if build or
                   { the source and the unit are at different locations }
                   (unit_path<>source_path) or
                   { the source is newer than the unit file }
                     (source_time>unit_time) then
                     begin
                        units:=new(punit,init(source_path,s,units));
                        units^.sourcename:=stringdup(s);
                        units^.sourceext:=stringdup(target_info.sourceext);
                        readunit:=nil;
                        exit;
                     end;
                end;
           end;
         assign(unitfile,unit_path+s+target_info.unitext);
         reset(unitfile,1);
         units:=new(punit,init(unit_path,s+target_info.unitext,units));

         unitfilesize:=filesize(unitfile);

         if unitfilesize>maxunitsize then
           fatalerror(error_reading_unit);

         { maybe write a message }
         if not(quiet) then
           writeln('Loading '+s+target_info.unitext);
         getmem(rbuffer,unitfilesize);
         blockread(unitfile,rbuffer^,unitfilesize,bufferl);
         close(unitfile);

         { check the format }
         for b:=0 to 5 do
           if rbuffer^[b]<>byte(unitheader[b]) then
             fatalerror(malformed_unit);

         { check the target OS }
         if (rbuffer^[8])<>byte(target_info.target) then
           fatalerror(not_same_target);

         bufferp:=sizeof(unitheader);

         units^.symtab:=new(punitsymtable,load(s));
         readunit:=units^.symtab;

         { read the checksum of the file }
         checksum:=plongint(@rbuffer^[10])^;
         units^.symtab^.maschstart:=plongint(@rbuffer^[16])^;
         freemem(rbuffer,unitfilesize);

         rbuffer:=oldbuffer;
         bufferp:=oldbufferp;
         bufferl:=oldbufferl;
      end;

begin
   units:=nil;
end.
